{-|
Module      : Data.Graph.Clustering.ILouvain
Description : Purely functional (Inductive) Louvain clustering
Copyright   : (c) Alexandre Delanoë, CNRS, 2020-Present
License     : AGPL + CECILL v3
Maintainer  : alexandre.delanoe+louvain@iscpif.fr
Stability   : experimental
Portability : POSIX

ILouvain: really inductive Graph

-}

{-# LANGUAGE ConstrainedClassMethods #-}
{-# LANGUAGE FlexibleInstances #-}
{-# LANGUAGE InstanceSigs      #-}
{-# LANGUAGE NoImplicitPrelude #-}

module Data.Graph.Clustering.ILouvain
  where

import Data.Maybe (catMaybes)
import Data.List (zip, cycle)
import Protolude hiding (empty, (&))
import Data.Graph.Inductive

------------------------------------------------------------------------
-- HyperGraph Definition
type HyperGraph   a b = Gr (Gr () a) b
type HyperContext a b = Context (Gr () a) b

-- TODO Later (hypothesis still)
-- type StreamGraph  a b = Gr a (Gr () b)
------------------------------------------------------------------------
-- Spoon Graph
--    1
--   / \
--  2   3
--   \ /
--    4
--    |
--    5

spoon :: HyperGraph Double Double
spoon = mkGraph ns es
  where
    ns :: [LNode (Gr () Double)]
    ns = zip [1..6] (cycle [empty])

    es :: [LEdge Double]
    es = [ (1, 2, 1.0)
         , (1, 3, 1.0)
         , (2, 4, 1.0)
         , (3, 4, 1.0)
         , (4, 5, 1.0)
         ]

-- | Needed functions (WIP)
-- mv: a Node elsewhere in the HyperGraph

-- Move Properties:
-- mv Node into another Node and remove it again should be equal to
-- identity

-- property test:
-- mv (mv g [a] [b]) [b,a] [] = identity
-- mv g [a] [] == g
-- mv g [a,b] [] /= g
-- mv (mv spoon [1] [4]) [4,1] [] = identity

-- Move target type
mv' :: HyperGraph a a
   -> [Node] -> [Node]
   -> HyperGraph a a
mv' g []  [] = g
mv' g [_] [] = g
mv' g (x:xs) [] = undefined

-- | Start simple (without path)
mv :: HyperGraph a a
   -> Node -> Node
   -> HyperGraph a a
mv g n1 n2 =  delNode n1 g
-- buildGr $ catMaybes [c1, c2]
  where
    (c1, g1) = match n1 g



{-
insertContext :: HyperContext a b
              -> HyperContext a b
              -> HyperContext a b
insertContext (a1,n,l,a2) (a1',n',l',a2') = (a1,n,l&l',a2)

--}

------------------------------------------------------------------------
-- | Recursive Node of Graph
{-
rnodes :: RGraph -> [Node]
rnodes Empty = []
rnodes g     = concat $ map (\(x1, x2) -> [x1] <> rnodes x2) $ labNodes g

rlabNodes :: Graph' a b -> [LNode a]
rlabNodes Empty' = []
rlabNodes g      = labNodes g
-}

------------------------------------------------------------------------
-- Paths in the Graph to be tested
-- Directed graph strategy
path_dir :: Graph gr => gr a b -> [[Node]]
path_dir g = map (\xs -> dfs xs g ) (components g)

-- UnDirected graph strategy
path :: (DynGraph gr, Eq b) => gr a b -> [[Node]]
path g' = map sortNodes cs
  where
    sortNodes ns = case head $ sortOn (Down . (deg g)) ns of
      Nothing -> []
      Just  n -> dfs [n] g  -- dfs for glustering, bfs for klustering
    cs = components g
    g  = undir g'
------------------------------------------------------------------------
------------------------------------------------------------------------
